var ts = require('typescript');
var extend = require('xtend');
var has = require('has');
var Promise = require('native-or-bluebird');
var os_1 = require('os');
var path_1 = require('path');
var fs_1 = require('../utils/fs');
var path_2 = require('../utils/path');
var references_1 = require('../utils/references');
var config_1 = require('../utils/config');
var typings_1 = require('../typings');
var SEPARATOR = '~';
var formattingOptions = {
    InsertSpaceAfterCommaDelimiter: true,
    InsertSpaceAfterSemicolonInForStatements: true,
    InsertSpaceBeforeAndAfterBinaryOperators: true,
    InsertSpaceAfterKeywordsInControlFlowStatements: true,
    InsertSpaceAfterFunctionKeywordForAnonymousFunctions: true,
    InsertSpaceAfterOpeningAndBeforeClosingNonemptyBrackets: false,
    InsertSpaceAfterOpeningAndBeforeClosingNonemptyParenthesis: false,
    PlaceOpenBraceOnNewLineForFunctions: false,
    PlaceOpenBraceOnNewLineForControlBlocks: false,
    IndentSize: 2,
    TabSize: 4,
    IndentStyle: ts.IndentStyle.Smart,
    NewLineCharacter: os_1.EOL,
    ConvertTabsToSpaces: true
};
function compile(tree, options) {
    var files = {};
    return Promise.all([
        compileDependencyTree(tree, extend(options, { browser: false, files: files })),
        compileDependencyTree(tree, extend(options, { browser: true, files: files }))
    ])
        .then(function (_a) {
        var main = _a[0], browser = _a[1];
        return ({ main: main, browser: browser });
    });
}
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = compile;
function getStringifyOptions(tree, options, parent) {
    var overrides = {};
    var isTypings = typeof tree.typings === 'string';
    var main = isTypings ? tree.typings : tree.main;
    var browser = isTypings ? tree.browserTypings : tree.browser;
    if (options.browser && browser) {
        if (typeof browser === 'string') {
            var mainDefinition = path_2.resolveFrom(tree.src, path_2.normalizeToDefinition(main));
            var browserDefinition = path_2.resolveFrom(tree.src, path_2.normalizeToDefinition(browser));
            overrides[mainDefinition] = browserDefinition;
        }
        else {
            var browserOverrides = browser;
            for (var _i = 0, _a = Object.keys(browserOverrides); _i < _a.length; _i++) {
                var key = _a[_i];
                if (!path_2.isModuleName(key)) {
                    var from = path_2.resolveFrom(tree.src, path_2.normalizeToDefinition(key));
                    var to = path_2.resolveFrom(tree.src, path_2.normalizeToDefinition(browserOverrides[key]));
                    overrides[from] = to;
                }
                else {
                    overrides[key] = browserOverrides[key];
                }
            }
        }
    }
    var imported = {};
    var referenced = {};
    var dependencies = {};
    var entry = main == null ? main : path_2.resolveFrom(tree.src, path_2.normalizeToDefinition(main));
    return extend(options, {
        tree: tree,
        entry: entry,
        isTypings: isTypings,
        overrides: overrides,
        imported: imported,
        referenced: referenced,
        dependencies: dependencies,
        parent: parent
    });
}
function compileDependencyTree(tree, options) {
    return compileDependencyPath(null, getStringifyOptions(tree, options, undefined));
}
function compileDependencyPath(path, options) {
    var tree = options.tree, entry = options.entry;
    if (tree.missing) {
        return Promise.reject(new Error("Missing dependency \"" + options.name + "\", unable to compile dependency tree"));
    }
    if (path == null) {
        if (entry == null) {
            return Promise.reject(new Error(("Unable to resolve entry .d.ts file for \"" + options.name + "\", ") +
                'please make sure the module has a main or typings field'));
        }
        return stringifyDependencyPath(path_2.resolveFrom(tree.src, entry), options);
    }
    return stringifyDependencyPath(path_2.resolveFrom(tree.src, path), options);
}
function cachedReadFileFrom(path, options) {
    if (!has(options.files, path)) {
        options.files[path] = fs_1.readFileFrom(path);
    }
    return options.files[path];
}
function cachedStringifyOptions(name, compileOptions, options) {
    var tree = getDependency(name, options);
    if (!has(options.dependencies, name)) {
        if (tree) {
            options.dependencies[name] = getStringifyOptions(tree, compileOptions, options);
        }
        else {
            options.dependencies[name] = null;
        }
    }
    return options.dependencies[name];
}
function getDependency(name, options) {
    var tree = options.tree, overrides = options.overrides;
    if (has(options.overrides, name)) {
        return tree.dependencies[overrides[name]];
    }
    if (has(tree.dependencies, name)) {
        return tree.dependencies[name];
    }
}
function stringifyDependencyPath(path, options) {
    var definitionPath = path_2.normalizeToDefinition(path);
    var tree = options.tree, ambient = options.ambient, cwd = options.cwd, browser = options.browser, name = options.name, files = options.files, meta = options.meta, entry = options.entry;
    if (has(options.overrides, definitionPath)) {
        definitionPath = options.overrides[definitionPath];
    }
    return cachedReadFileFrom(definitionPath, options)
        .then(function (rawContents) {
        var info = ts.preProcessFile(rawContents);
        var ambientModules = info.ambientExternalModules || [];
        if (info.isLibFile) {
            return;
        }
        if (ambientModules.length && !ambient) {
            return Promise.reject(new TypeError(("Attempted to compile " + options.name + " as a ") +
                "dependency, but it contains ambient module declarations. Did " +
                "you want to specify \"--ambient\" instead?"));
        }
        var importedFiles = info.importedFiles.map(function (x) { return path_2.isModuleName(x.fileName) ? x.fileName : path_2.resolveFrom(path, x.fileName); });
        if (ambient) {
            Object.keys(tree.dependencies).forEach(function (x) { return importedFiles.push(x); });
        }
        var imports = importedFiles.map(function (path) {
            if (has(options.imported, path)) {
                return;
            }
            if (ambientModules.indexOf(path) > -1) {
                return;
            }
            options.imported[path] = true;
            if (path_2.isModuleName(path)) {
                var _a = getModuleNameParts(path), dependencyName = _a[0], dependencyPath = _a[1];
                var moduleName = ambient ? dependencyName : "" + name + SEPARATOR + dependencyName;
                var compileOptions = { cwd: cwd, browser: browser, files: files, name: moduleName, ambient: false, meta: meta };
                var stringifyOptions = cachedStringifyOptions(dependencyName, compileOptions, options);
                if (!stringifyOptions) {
                    return;
                }
                return compileDependencyPath(dependencyPath, stringifyOptions);
            }
            return stringifyDependencyPath(path, options);
        });
        return Promise.all(imports)
            .then(function (files) {
            var stringifyOptions = extend(options, { originalPath: path });
            var contents = rawContents.replace(references_1.REFERENCE_REGEXP, '');
            files.push(stringifyFile(definitionPath, contents, stringifyOptions));
            return files.filter(function (x) { return x != null; }).join(os_1.EOL + os_1.EOL);
        });
    }, function () {
        var authorPhrase = options.parent ? "The author of \"" + options.parent.name + "\" needs" : 'You need';
        if (path === entry) {
            return Promise.reject(new Error(("Unable to read typings in \"" + options.name + "\". ") +
                (authorPhrase + " to add an entry to \"" + config_1.CONFIG_FILE + "\" with the typings")));
        }
        return Promise.reject(new Error(("Unable to read " + path_2.relativeTo(tree.src, path) + " from \"" + options.name + "\". ") +
            (authorPhrase + " to check that the entry in \"" + config_1.CONFIG_FILE + "\" is ") +
            ("complete (the path \"" + path_2.relativeTo(tree.src, path) + "\" appears to be missing)")));
    });
}
function getModuleNameParts(moduleName) {
    var parts = moduleName.split(/[\\\/]/);
    var dependencyName = parts.shift();
    var dependencyPath = parts.length === 0 ? null : parts.join('/');
    return [dependencyName, dependencyPath];
}
function format(text) {
    var sourceFile = ts.createSourceFile('file.ts', text, 2, true);
    var ruleProvider = getRuleProvider(formattingOptions);
    var edits = ts.formatting.formatDocument(sourceFile, ruleProvider, formattingOptions);
    return applyEdits(text, edits);
}
function applyEdits(text, edits) {
    var result = text;
    var length = edits.length;
    while (length--) {
        var change = edits[length];
        var head = result.slice(0, change.span.start);
        var tail = result.slice(change.span.start + change.span.length);
        result = head + change.newText + tail;
    }
    return result;
}
function getRuleProvider(options) {
    var ruleProvider = new ts.formatting.RulesProvider();
    ruleProvider.ensureUpToDate(options);
    return ruleProvider;
}
function stringifyFile(path, contents, options) {
    var sourceFile = ts.createSourceFile(path, contents, 2, true);
    var tree = options.tree, name = options.name, originalPath = options.originalPath;
    var source = path_2.isHttp(path) ? path : path_1.relative(options.cwd, path);
    var prefix = options.meta ? "// Compiled using " + config_1.PROJECT_NAME + "@" + typings_1.VERSION + os_1.EOL + "// Source: " + source + os_1.EOL : '';
    if (options.ambient) {
        if (sourceFile.externalModuleIndicator) {
            throw new TypeError(("Attempted to compile " + options.name + " as an ambient ") +
                "module declaration, but it has external module indicators. Did you " +
                "want to omit \"--ambient\"?");
        }
        return format(prefix + contents.trim());
    }
    var isES6Export = true;
    var wasDeclared = false;
    function importPath(name) {
        if (path_2.isModuleName(name)) {
            var moduleName_1 = getModuleNameParts(name)[0];
            if (options.dependencies[moduleName_1] == null) {
                return name;
            }
            return "" + options.name + SEPARATOR + name;
        }
        var relativePath = path_2.relativeTo(tree.src, path_2.resolveFrom(path, name));
        return path_2.normalizeSlashes(path_1.join(options.name, relativePath));
    }
    function replacer(node) {
        if (node.kind === 227) {
            isES6Export = !node.isExportEquals;
        }
        if (node.kind === 9 &&
            (node.parent.kind === 228 ||
                node.parent.kind === 222)) {
            return " '" + importPath(node.text) + "'";
        }
        if (node.kind === 122) {
            wasDeclared = true;
            return sourceFile.text.slice(node.getFullStart(), node.getStart());
        }
        if (node.kind === 232) {
            var path_3 = importPath(node.expression.text);
            return " require('" + path_3 + "')";
        }
    }
    function read(start, end) {
        var text = sourceFile.text.slice(start, end);
        if (start === 0) {
            return text.replace(/^\s+$/, '');
        }
        if (end == null) {
            return text.replace(/\s+$/, '');
        }
        if (wasDeclared) {
            wasDeclared = false;
            return text.replace(/^\s+/, '');
        }
        return text;
    }
    var moduleText = processTree(sourceFile, replacer, read);
    var isEntry = originalPath === options.entry;
    if (isEntry && options.isTypings) {
        return format(prefix + declareText(name, moduleText));
    }
    var moduleName = name + "/" + path_2.normalizeSlashes(path_2.relativeTo(tree.src, path_2.fromDefinition(path)));
    var declared = declareText(moduleName, moduleText);
    if (!isEntry) {
        return format(prefix + declared);
    }
    var importText = isES6Export ?
        "export * from '" + moduleName + "';" :
        "import main = require('" + moduleName + "');" + os_1.EOL + "export = main;";
    return format(prefix + declared + os_1.EOL + declareText(name, importText));
}
function declareText(name, text) {
    return "declare module '" + name + "' {" + os_1.EOL + text + os_1.EOL + "}";
}
function processTree(sourceFile, replacer, reader) {
    var code = '';
    var position = 0;
    function skip(node) {
        position = node.end;
    }
    function readThrough(node) {
        code += reader(position, node.pos);
        position = node.pos;
    }
    function visit(node) {
        readThrough(node);
        var replacement = replacer(node);
        if (replacement != null) {
            code += replacement;
            skip(node);
        }
        else {
            ts.forEachChild(node, visit);
        }
    }
    visit(sourceFile);
    code += reader(position);
    return code;
}
//# sourceMappingURL=compile.js.map