var fs = require('graceful-fs');
var thenify = require('thenify');
var stripBom = require('strip-bom');
var parse = require('parse-json');
var popsicle = require('popsicle');
var popsicleCache = require('popsicle-cache');
var popsicleStatus = require('popsicle-status');
var detectIndent = require('detect-indent');
var sortKeys = require('sort-keys');
var mdp = require('mkdirp');
var uniq = require('array-uniq');
var Promise = require('native-or-bluebird');
var lockfile = require('lockfile');
var promise_finally_1 = require('promise-finally');
var path_1 = require('path');
var config_1 = require('./config');
var path_2 = require('./path');
var references_1 = require('./references');
var requestFileCache = popsicleCache({
    store: new popsicleCache.Store({ path: path_1.join(config_1.CACHE_DIR, 'http') })
});
var mainTypingsDir = path_1.join(config_1.TYPINGS_DIR, 'main/definitions');
var browserTypingsDir = path_1.join(config_1.TYPINGS_DIR, 'browser/definitions');
var ambientMainTypingsDir = path_1.join(config_1.TYPINGS_DIR, 'main/ambient');
var ambientBrowserTypingsDir = path_1.join(config_1.TYPINGS_DIR, 'browser/ambient');
exports.stat = thenify(fs.stat);
exports.readFile = thenify(fs.readFile);
exports.writeFile = thenify(fs.writeFile);
exports.mkdirp = thenify(mdp);
exports.unlink = thenify(fs.unlink);
exports.lock = thenify(lockfile.lock);
exports.unlock = thenify(lockfile.unlock);
function isFile(path) {
    return exports.stat(path).then(function (stat) { return stat.isFile(); }, function () { return false; });
}
exports.isFile = isFile;
function readJson(path) {
    return exports.readFile(path, 'utf8')
        .then(stripBom)
        .then(function (contents) { return parseJson(contents, path); });
}
exports.readJson = readJson;
function writeJson(path, json, indent) {
    if (indent === void 0) { indent = 2; }
    return exports.writeFile(path, JSON.stringify(json, null, indent));
}
exports.writeJson = writeJson;
function readConfig(path) {
    return readJson(path).then(function (data) { return parseConfig(data, path); });
}
exports.readConfig = readConfig;
function readConfigFrom(path) {
    return readJsonFrom(path).then(function (data) { return parseConfig(data, path); });
}
exports.readConfigFrom = readConfigFrom;
function parseConfig(config, path) {
    return config;
}
exports.parseConfig = parseConfig;
function readHttp(url) {
    return popsicle.get({
        url: url,
        use: [
            popsicle.plugins.headers,
            popsicle.plugins.unzip,
            popsicle.plugins.concatStream('string')
        ]
    })
        .use(requestFileCache)
        .use(popsicleStatus(200))
        .then(function (x) { return x.body; });
}
exports.readHttp = readHttp;
function readFileFrom(from) {
    return path_2.isHttp(from) ? readHttp(from) : exports.readFile(from, 'utf8');
}
exports.readFileFrom = readFileFrom;
function readJsonFrom(from) {
    return readFileFrom(from)
        .then(stripBom)
        .then(function (contents) { return parseJson(contents, from); });
}
exports.readJsonFrom = readJsonFrom;
function parseJson(contents, path) {
    return parse(contents, null, path);
}
exports.parseJson = parseJson;
function transformFile(path, transform) {
    function handle(contents) {
        return Promise.resolve(transform(contents))
            .then(function (contents) { return exports.writeFile(path, contents); });
    }
    var lockfile = path + ".lock";
    var lockOptions = { wait: 250, retries: 25, stale: 60000 };
    var result = exports.lock(lockfile, lockOptions)
        .then(function () {
        return exports.readFile(path, 'utf8');
    })
        .then(function (contents) { return handle(contents); }, function () { return handle(undefined); });
    return promise_finally_1.default(result, function () { return exports.unlock(lockfile); });
}
exports.transformFile = transformFile;
function transformJson(path, transform) {
    return transformFile(path, function (contents) {
        var indent = contents ? detectIndent(contents).indent : 2;
        var json = contents ? parseJson(contents, path) : undefined;
        return Promise.resolve(transform(json))
            .then(function (json) { return JSON.stringify(json, null, indent || 2); });
    });
}
exports.transformJson = transformJson;
function transformConfig(cwd, transform) {
    var path = path_1.join(cwd, config_1.CONFIG_FILE);
    return transformJson(path, function (config) {
        if (config === void 0) { config = {}; }
        return Promise.resolve(transform(parseConfig(config, path)))
            .then(function (config) {
            if (config.dependencies) {
                config.dependencies = sortKeys(config.dependencies);
            }
            if (config.devDependencies) {
                config.devDependencies = sortKeys(config.devDependencies);
            }
            if (config.ambientDependencies) {
                config.ambientDependencies = sortKeys(config.ambientDependencies);
            }
            if (config.ambientDevDependencies) {
                config.ambientDevDependencies = sortKeys(config.ambientDevDependencies);
            }
            return config;
        });
    });
}
exports.transformConfig = transformConfig;
function transformDtsFile(path, transform) {
    var cwd = path_1.dirname(path);
    return transformFile(path, function (contents) {
        var typings = references_1.parseReferences(contents, cwd);
        return Promise.resolve(transform(typings))
            .then(function (typings) { return references_1.stringifyReferences(uniq(typings).sort(), cwd); });
    });
}
exports.transformDtsFile = transformDtsFile;
function writeDependency(contents, options) {
    var _a = getDependencyLocation(options), mainFile = _a.mainFile, browserFile = _a.browserFile, mainDtsFile = _a.mainDtsFile, browserDtsFile = _a.browserDtsFile;
    return Promise.all([
        exports.mkdirp(path_1.dirname(mainFile)),
        exports.mkdirp(path_1.dirname(browserFile))
    ])
        .then(function () {
        return Promise.all([
            exports.writeFile(mainFile, contents.main || ''),
            exports.writeFile(browserFile, contents.browser || '')
        ]);
    })
        .then(function () {
        return Promise.all([
            transformDtsFile(mainDtsFile, function (typings) { return typings.concat([mainFile]); }),
            transformDtsFile(browserDtsFile, function (typings) { return typings.concat([browserFile]); })
        ]);
    })
        .then(function () { return undefined; });
}
exports.writeDependency = writeDependency;
function removeDependency(options) {
    var _a = getDependencyLocation(options), mainFile = _a.mainFile, browserFile = _a.browserFile, mainDtsFile = _a.mainDtsFile, browserDtsFile = _a.browserDtsFile;
    return Promise.all([
        exports.unlink(mainFile).catch(function () { return false; }),
        exports.unlink(browserFile).catch(function () { return false; })
    ])
        .then(function () {
        return Promise.all([
            transformDtsFile(mainDtsFile, function (typings) { return typings.filter(function (x) { return x !== mainFile; }); }),
            transformDtsFile(browserDtsFile, function (typings) { return typings.filter(function (x) { return x !== browserFile; }); })
        ]);
    })
        .then(function () { return undefined; });
}
exports.removeDependency = removeDependency;
function getDependencyLocation(options) {
    var typingsDir = path_1.join(options.cwd, config_1.TYPINGS_DIR);
    var mainDtsFile = path_1.join(typingsDir, config_1.DTS_MAIN_FILE);
    var browserDtsFile = path_1.join(typingsDir, config_1.DTS_BROWSER_FILE);
    var mainDir = options.ambient ? ambientMainTypingsDir : mainTypingsDir;
    var browserDir = options.ambient ? ambientBrowserTypingsDir : browserTypingsDir;
    var mainFile = path_1.join(options.cwd, mainDir, options.name, path_2.toDefinition(options.name));
    var browserFile = path_1.join(options.cwd, browserDir, options.name, path_2.toDefinition(options.name));
    return { mainFile: mainFile, browserFile: browserFile, mainDtsFile: mainDtsFile, browserDtsFile: browserDtsFile };
}
//# sourceMappingURL=fs.js.map